

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Sum                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Sum : public Reduce_<T, T> {
public :
  Sum(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    *res += *((T*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE)
       blockSum(res, srcBase + srcOffset, steps, exts, rank) ;
    else 
       oc_blockSum(res, srcOffset, rank, exts, steps, vol, 
                      elementLen, srcBase, status, OC_SUM) ;
  }

} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class SumMsk                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SumMsk : public Reduce2_<T, T, int> {
public :
  SumMsk(const DAD* src, const DAD* msk) : Reduce2_<T, T, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res += *((T*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    if (status == IN_CORE)
       blockSum(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank) ;
    else 
    {
//  AdlibPrintf("SumMsk::exts[0]=%d,exts[1]=%d,exts[2]=%d\n",exts[0],exts[1],exts[2]);
//  AdlibPrintf("SumMsk::steps1[0]=%d,steps1[1]=%d,steps1[2]=%d\n",steps1[0],steps1[1],steps1[2]);
//  AdlibPrintf("SumMsk::steps2[0]=%d,steps2[1]=%d,steps2[2]=%d\n",steps2[0],steps2[1],steps2[2]);
//  AdlibPrintf("SumMsk::srcOffset=%d,mskOffset=%d,rank=%d,vol=%d\n",srcOffset,mskOffset,rank,vol);
//  AdlibPrintf("SumMsk::elementLen1=%d,elementLen2=%d\n",elementLen1,elementLen2);
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blockSum(res, srcOffset, steps1, mskOffset, steps2, rank, exts, vol, 
                  elementLen1, elementLen2, src1Base, src2Base, status, OC_SUM) ;
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Product                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Product : public Reduce_<T, T> {
public :
  Product(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    *res *= *((T*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset) {
    blockProduct(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ProductMsk                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ProductMsk : public Reduce2_<T, T, int> {
public :
  ProductMsk(const DAD* src, const DAD* msk) : Reduce2_<T, T, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res *= *((T*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockProduct(res, src1Base + srcOffset, steps1,
                      src2Base + mskOffset, steps2, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Maxval                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Maxval : public Reduce_<T, T> {
public :
  Maxval(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) > *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostneg(res) ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE)
      blockMaxval(res, srcBase + srcOffset, steps, exts, rank) ;
    else 
       oc_blockReduce(res, srcOffset, rank, exts, steps, vol, 
                      elementLen, srcBase, status, OC_MAXVAL) ;
  }

} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxvalMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxvalMsk : public Reduce2_<T, T, int> {
public :
  MaxvalMsk(const DAD* src, const DAD* msk) :
                Reduce2_<T, T, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) > *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostneg(res) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    if (status == IN_CORE)
       blockMaxval(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank) ;
    else { 
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blockReduce(res, srcOffset, steps1, mskOffset, steps2, rank, exts, vol, 
            elementLen1, elementLen2, src1Base, src2Base, status, OC_MAXVAL) ; 
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
     }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Minval                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Minval : public Reduce_<T, T> {
public :
  Minval(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) < *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostpos(res) ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE)
      blockMinval(res, srcBase + srcOffset, steps, exts, rank) ;
    else 
       oc_blockReduce(res, srcOffset, rank, exts, steps, vol, 
                      elementLen, srcBase, status, OC_MINVAL) ;
  }

} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinvalMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinvalMsk : public Reduce2_<T, T, int> {
public :
  MinvalMsk(const DAD* _src, const DAD* _msk) : Reduce2_<T, T, int>(_src, _msk) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) < *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostpos(res) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    if (status == IN_CORE)
        blockMinval(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank) ;
    else {
        int old_dfsMode = defaultStorageMode ;
        defaultStorageMode = COLUMN_MAJOR ;
        PFILE* PFile1 = (PFILE *)src1Base ;
        int old_sMode1 = PFile1->share->storage;
        PFILE* PFile2 = (PFILE *)src2Base ;   
        int old_sMode2 = PFile2->share->storage;
        if (old_sMode1 != COLUMN_MAJOR) {
          PFile1->share->storage = COLUMN_MAJOR ;
          if (old_sMode2 == COLUMN_MAJOR) 
                PFile2->share->storage = ROW_MAJOR ;
          else
                PFile2->share->storage = COLUMN_MAJOR ;
        }

        oc_blockReduce(res, srcOffset, steps1, mskOffset, steps2, rank, exts, vol, 
            elementLen1, elementLen2, src1Base, src2Base, status, OC_MINVAL) ; 
        PFile1->share->storage = old_sMode1 ;
        PFile2->share->storage = old_sMode2 ;
        defaultStorageMode = old_dfsMode ;
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class All                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class All : public Reduce_<int, int> {
public :
  All(const DAD* _src) : Reduce_<int, int>(_src) {}

private :
  void opCombine(void* buf) {
    *res = *res && *((int*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset) {
   if (status == IN_CORE)
    blockAnd(res, srcBase + srcOffset, steps, exts, rank) ;
   else oc_blocklogical(res, srcOffset, rank, exts, steps, vol, 
                      elementLen, srcBase, status, OC_ALL) ;

  }
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AllMsk                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllMsk : public Reduce2_<int, int, int> {
public :
  AllMsk(const DAD* src, const DAD* msk) : Reduce2_<int, int, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res = *res && *((int*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
//xh
    if (status == IN_CORE)
       blocklogical(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank,ALL) ;
    else 
    {
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blocklogical(res, srcOffset, steps1, mskOffset, steps2, rank, exts, vol, 
                  elementLen1, elementLen2, src1Base, src2Base, status, OC_ALL) ;
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
    }
  }
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Any                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Any : public Reduce_<int, int> {
public :
  Any(const DAD* _src) : Reduce_<int, int>(_src) {}

private :
  void opCombine(void* buf) {
    *res = *res || *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE)
      blockOr(res, srcBase + srcOffset, steps, exts, rank) ;
    else oc_blocklogical(res, srcOffset, rank, exts, steps, vol, 
                      elementLen, srcBase, status, OC_ANY) ;

  }
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AnyMsk                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AnyMsk : public Reduce2_<int, int, int> {
public :
  AnyMsk(const DAD* src, const DAD* msk) : Reduce2_<int, int, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res = *res || *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
//xh
    if (status == IN_CORE)
       blocklogical(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank,ANY) ;
    else 
    {
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blocklogical(res, srcOffset, steps1, mskOffset, steps2, rank, exts, vol, 
                  elementLen1, elementLen2, src1Base, src2Base, status, OC_ANY) ;
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
    }
  }
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Count                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Count : public Reduce_<int, int> {
public :
  Count(const DAD* _src) : Reduce_<int, int>(_src) {}

private :
  void opCombine(void* buf) {
//    if(*((int*) buf)) (*res)++ ;
    *res += *((int*) buf) ;              // modified by Zheng Gengbin
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE)
      blockCount(res, srcBase + srcOffset, steps, exts, rank) ;
    else  oc_blocklogical(res, srcOffset, rank, exts, steps, vol, 
                      elementLen, srcBase, status, OC_COUNT) ;

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class CountMsk                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CountMsk : public Reduce2_<int, int, int> {
public :
  CountMsk(const DAD* src, const DAD* msk) : Reduce2_<int, int, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res += *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    if (status == IN_CORE)
       blocklogical(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank,COUNT) ;
    else 
    {
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blocklogical(res, srcOffset, steps1, mskOffset, steps2, rank, exts, vol, 
                  elementLen1, elementLen2, src1Base, src2Base, status, OC_COUNT) ;
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
    }
  }
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class DotProduct                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
class DotProduct : public Reduce2_<S, T, U> {
public :
  DotProduct(const DAD* _src1, const DAD* _src2) :
                 Reduce2_<S, T, U>(_src1, _src2) {}

private :
  void opCombine(void* buf) {
    *res += *((S*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int src1Offset, const int src2Offset) {
    if (status == IN_CORE)
       blockDotProduct(res, src1Base + src1Offset, steps1,
                         src2Base + src2Offset, steps2, exts, rank) ;
    else {
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blockDotProduct(res, src1Offset, src2Offset, 
          rank, exts, steps1, steps2, vol, 
          elementLen1, elementLen2, src1Base, src2Base, status) ;
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
    }
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class BoolDotProduct                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BoolDotProduct : public Reduce2_<int, int, int> {
public :
  BoolDotProduct(const DAD* _src1, const DAD* _src2) :
        Reduce2_<int, int, int>(_src1, _src2) {}

private :
  void opCombine(void* buf) {
    *res += *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int src1Offset, const int src2Offset) {
    if (status == IN_CORE)
      blockBoolDotProduct(res, src1Base + src1Offset, steps1,
                             src2Base + src2Offset, steps2, exts, rank) ;
    else {
       int old_dfsMode = defaultStorageMode ;
       defaultStorageMode = COLUMN_MAJOR ;
       PFILE* PFile1 = (PFILE *)src1Base ;
       int old_sMode1 = PFile1->share->storage;
       PFILE* PFile2 = (PFILE *)src2Base ;   
       int old_sMode2 = PFile2->share->storage;
       if (old_sMode1 != COLUMN_MAJOR) {
         PFile1->share->storage = COLUMN_MAJOR ;
         if (old_sMode2 == COLUMN_MAJOR) 
               PFile2->share->storage = ROW_MAJOR ;
         else
               PFile2->share->storage = COLUMN_MAJOR ;
       }

       oc_blockBoolDotProduct(res, src1Offset, src2Offset, 
          rank, exts, steps1, steps2, vol, 
          elementLen1, elementLen2, src1Base, src2Base, status) ;
       PFile1->share->storage = old_sMode1 ;
       PFile2->share->storage = old_sMode2 ;
       defaultStorageMode = old_dfsMode ;
    }
  }
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Maxloc                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Maxloc : public Locate_<T, T> {
public :
  Maxloc(const DAD* _src) : Locate_<T, T>(_src) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) < *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostneg(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE) {
       blockMaxloc(val, tags, rank,
                srcBase + srcOffset, srcSteps, blockBases, blockSteps,
                exts, rank) ;
       if(major != COLUMN_MAJOR) {
         int tmp_tags[10], i;
         for (i=0; i<rank; i++)
           tmp_tags[i] = tags[i];
         for (i=0; i<rank; i++)
           tags[i] = tmp_tags[rank-1-i];
       }
    }
    else
    {

       oc_blockMaxloc(val, tags, rank, srcBase, 
                srcOffset, srcSteps, blockBases, blockSteps, status, 
                exts, rank, vol, elementLen) ;

//  AdlibPrintf("Reduce::val=%d, tags[0]=%d,tags[1]=%d\n",*val,tags[0],tags[1]);
    }

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxlocMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxlocMsk : public Locate2_<T, T, int> {
public :
  MaxlocMsk(const DAD* _src, const DAD* _msk) :
                Locate2_<T, T, int>(_src, _msk) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) < *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostneg(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    if (status == IN_CORE) {
       blockMaxloc(val, tags, rank,
                src1Base + srcOffset, src1Steps,
                src2Base + mskOffset, src2Steps,
                blockBases, blockSteps,
                exts, rank) ;
       if(src1_major != COLUMN_MAJOR) {
         int tmp_tags[10], i;
         for (i=0; i<rank; i++)
           tmp_tags[i] = tags[i];
         for (i=0; i<rank; i++)
           tags[i] = tmp_tags[rank-1-i];
       }
    }
    else 
       oc_blockMaxloc(val, tags, rank,
          src1Base, srcOffset, src1Steps,
          src2Base, mskOffset, src2Steps,
          blockBases, blockSteps, status, 
          exts, rank, vol, elementLen1, elementLen2) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Minloc                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Minloc : public Locate_<T, T> {
public :
  Minloc(const DAD* _src) : Locate_<T, T>(_src) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) > *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostpos(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset) {
    if (status == IN_CORE) {
       blockMinloc(val, tags, rank,
                srcBase + srcOffset, srcSteps, blockBases, blockSteps,
                exts, rank) ;
       if(major != COLUMN_MAJOR) {
         int tmp_tags[10], i;
         for (i=0; i<rank; i++)
           tmp_tags[i] = tags[i];
         for (i=0; i<rank; i++)
           tags[i] = tmp_tags[rank-1-i];
       }
    }
    else
       oc_blockMinloc(val, tags, rank, srcBase, 
                srcOffset, srcSteps, blockBases, blockSteps, status, 
                exts, rank, vol, elementLen) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinlocMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinlocMsk : public Locate2_<T, T, int> {
public :
  MinlocMsk(const DAD* _src, const DAD* _msk) :
      Locate2_<T, T, int>(_src, _msk) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) > *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostpos(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    if (status == IN_CORE) {
       blockMinloc(val, tags, rank,
                src1Base + srcOffset, src1Steps,
                src2Base + mskOffset, src2Steps,
                blockBases, blockSteps,
                exts, rank) ;
       if(src1_major != COLUMN_MAJOR) {
         int tmp_tags[10], i;
         for (i=0; i<rank; i++)
           tmp_tags[i] = tags[i];
         for (i=0; i<rank; i++)
           tags[i] = tmp_tags[rank-1-i];
       }
    }
    else 
       oc_blockMinloc(val, tags, rank,
          src1Base, srcOffset, src1Steps,
          src2Base, mskOffset, src2Steps,
          blockBases, blockSteps, status, 
          exts, rank, vol, elementLen1, elementLen2) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class SumDim                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SumDim : public ReduceDim_<T, T> {
public :
  SumDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
AdlibPrintf("resB %d\n",resBlock[MBLK_STA]);
    if (resBlock[MBLK_STA] == IN_CORE)
      blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
    else {
AdlibPrintf("Come to Combine %d \n",resBlock[MBLK_SIZ]);
       PFILE* Pfile = (PFILE*)resBase;
       oc_blockScatterSum(Pfile,resBlock,resExts,rank,(T*)buf,reselementLen,srcelementLen,OC_SUM);
       }
 
  }

  void opZero(const int resOffset) {
    if(status==IN_CORE)
      blockZero(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
  if(status==IN_CORE)
    blockVecSum(resBase + resOffset, resSteps,
		srcBase + srcOffset, srcSteps, resexts, resRank) ;
  else 
   oc_blockVecSum(srcBase,resBase,resOffset,srcOffset,srcexts,resexts,srcSteps,resSteps,
                     srcRank,resRank,srcvol,resvol,srcelementLen,reselementLen,
                     status,srcmajor,OC_SUM); 
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class SumDimMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SumDimMsk : public ReduceDim2_<T, T, int> {
public :
  SumDimMsk(const DAD* _res,
            const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
//xh
AdlibPrintf("SumDimMask %d %d %d\n",resBlock[MBLK_STA],IN_CORE,OUT_OF_CORE);
    if (resBlock[MBLK_STA] == IN_CORE) 
      blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
    else {
      PFILE* Pfile = (PFILE*)resBase;
      oc_blockScatterSum(Pfile,resBlock,resExts,rank,(T*)buf,reselementLen,srcelementLen,OC_SUM);
    }

  }

  void opZero(const int resOffset) {
   if (status == IN_CORE)
    blockZero(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset,
                const int mskOffset) {
//xh
   if (status == IN_CORE)
      blockVecSum(resBase + resOffset, resSteps,
		src1Base + srcOffset, src1Steps,
		src2Base + mskOffset, src2Steps, srcexts, resRank) ;
   else
       oc_blockVecSumMsk(src1Base,src2Base,resBase,
                      srcOffset,mskOffset,resOffset,
                      srcexts,resexts, 
                      src1Steps,src2Steps,resSteps,
                      srcRank,resRank,
                      srcvol,resvol,
                      srcelementLen,mskelementLen,reselementLen,
                      status,srcmajor,OC_SUM);

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ProductDim                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ProductDim : public ReduceDim_<T, T> {
public :
  ProductDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterProduct(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockOne(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecProduct(resBase + resOffset, resSteps,
		    srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ProductDimMsk                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ProductDimMsk : public ReduceDim2_<T, T, int> {
public :
  ProductDimMsk(const DAD* _res,
                const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterProduct(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockOne(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset, const int mskOffset) {
    blockVecProduct(resBase + resOffset, resSteps,
		    src1Base + srcOffset, src1Steps,
		    src2Base + mskOffset, src2Steps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxvalDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxvalDim : public ReduceDim_<T, T> {
public :
  MaxvalDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    if (resBlock[MBLK_STA] == IN_CORE)
      blockScatterMaxval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
    else {
         PFILE* Pfile = (PFILE*)resBase;
         oc_blockScatterVal(Pfile,resBlock,resExts,rank,(T*)buf,reselementLen,srcelementLen,OC_MAXVAL);
     }
  }

  void opZero(const int resOffset) {
    if (status == IN_CORE)
      blockMostneg(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    if (status == IN_CORE)
      blockVecMaxval(resBase + resOffset, resSteps,
		   srcBase + srcOffset, srcSteps, srcexts, resRank) ;
   else  oc_blockVecVal(srcBase,resBase,resOffset,srcOffset,
                       srcexts,resexts,srcSteps,resSteps,
                       srcRank,resRank,srcvol,resvol,
                       srcelementLen,reselementLen,
                       status,srcmajor,OC_MAXVAL);

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxvalDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxvalDimMsk : public ReduceDim2_<T, T, int> {
public :
  MaxvalDimMsk(const DAD* _res,
               const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    if (resBlock[MBLK_STA] == IN_CORE)
      blockScatterMaxval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
    else {
       PFILE* Pfile = (PFILE*)resBase;
       oc_blockScatterVal(Pfile,resBlock,resExts,rank,(T*)buf,reselementLen,srcelementLen,OC_MAXVAL);
     }

  }

  void opZero(const int resOffset) {
    if(status == IN_CORE)
      blockMostneg(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset, const int mskOffset) {
    if(status == IN_CORE)
      blockVecMaxval(resBase + resOffset, resSteps,
		   src1Base + srcOffset, src1Steps,
		   src2Base + mskOffset, src2Steps, srcexts, resRank) ;
    else   oc_blockVecValMsk(src1Base,src2Base,resBase,
                      srcOffset,mskOffset,resOffset,
                      srcexts,resexts, 
                      src1Steps,src2Steps,resSteps,
                      srcRank,resRank,
                      srcvol,resvol,
                      srcelementLen,mskelementLen,reselementLen,
                      status,srcmajor,OC_MAXVAL);
 
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinvalDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinvalDim : public ReduceDim_<T, T> {
public :
  MinvalDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    if(resBlock[MBLK_STA] == IN_CORE)
      blockScatterMinval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
    else {
          PFILE* Pfile = (PFILE*)resBase;
          oc_blockScatterVal(Pfile,resBlock,resExts,rank,(T*)buf,reselementLen,srcelementLen,OC_MINVAL);
        }

  }

  void opZero(const int resOffset) {
   if (status == IN_CORE)
    blockMostpos(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
   if (status == IN_CORE)
    blockVecMinval(resBase + resOffset, resSteps,
		   srcBase + srcOffset, srcSteps, srcexts, resRank) ;
   else oc_blockVecVal(srcBase,resBase,resOffset,srcOffset,
                        srcexts,resexts,srcSteps,resSteps,
                        srcRank,resRank,srcvol,resvol,
                        srcelementLen,reselementLen,
                        status,srcmajor,OC_MINVAL);
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinvalDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinvalDimMsk : public ReduceDim2_<T, T, int> {
public :
  MinvalDimMsk(const DAD* _res,
               const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    if (resBlock[MBLK_STA] == IN_CORE)
      blockScatterMinval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
    else  {
        PFILE* Pfile = (PFILE*)resBase;
        oc_blockScatterVal(Pfile,resBlock,resExts,rank,(T*)buf,reselementLen,srcelementLen,OC_MINVAL);
    }

  }

  void opZero(const int resOffset) {
   if (status == IN_CORE)
    blockMostpos(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset, const int mskOffset) {
    if (status == IN_CORE)
     blockVecMinval(resBase + resOffset, resSteps,
		   src1Base + srcOffset, src1Steps,
		   src2Base + mskOffset, src2Steps, srcexts, resRank) ;
    else   oc_blockVecValMsk(src1Base,src2Base,resBase,
                      srcOffset,mskOffset,resOffset,
                      srcexts,resexts, 
                      src1Steps,src2Steps,resSteps,
                      srcRank,resRank,
                      srcvol,resvol,
                      srcelementLen,mskelementLen,reselementLen,
                      status,srcmajor,OC_MINVAL);

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AllDim                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllDim : public ReduceDim_<int, int> {
public :
  AllDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<int, int>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
  if (resBlock[MBLK_STA] == IN_CORE)
    blockScatterAnd(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
  else {
         PFILE* Pfile = (PFILE*) resBase;
         oc_blockScatterlogical(Pfile,resBlock,resExts,rank,
                       (int*)buf,reselementLen,srcelementLen,OC_ALL);
        }

}

  void opZero(const int resOffset) {
  if (status == IN_CORE)
    blockOne(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
  if (status == IN_CORE)
    blockVecAnd(resBase + resOffset, resSteps,
		srcBase + srcOffset, srcSteps, srcexts, resRank) ;
  else  oc_blockVeclogical(srcBase,resBase,resOffset,srcOffset,
                     srcexts,resexts,srcSteps,resSteps,
                     srcRank,resRank,srcvol,resvol,srcelementLen,reselementLen,
                     status,srcmajor,OC_ALL); 
  }
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AllDimMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllDimMsk : public ReduceDim2_<int, int, int> {
public :
  AllDimMsk(const DAD* _res,
            const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<int, int, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
//xh
    if (resBlock[MBLK_STA] == IN_CORE) 
      blockScatterAnd(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
    else {
      PFILE* Pfile = (PFILE*)resBase;
      oc_blockScatterlogical(Pfile,resBlock,resExts,rank,(int*)buf,reselementLen,srcelementLen,OC_ALL);
    }
  }

  void opZero(const int resOffset) {
//xh
   if (status == IN_CORE)
      blockOne(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset,
                const int mskOffset) {
//xh
   if (status == IN_CORE)
       blockVecAnd(resBase + resOffset, resSteps,
		src1Base + srcOffset, src1Steps,
		src2Base + mskOffset, src2Steps, srcexts, resRank) ;
   else
       oc_blockVeclogicalMsk(src1Base,src2Base,resBase,
                      srcOffset,mskOffset,resOffset,
                      srcexts,resexts, 
                      src1Steps,src2Steps,resSteps,
                      srcRank,resRank,
                      srcvol,resvol,
                      srcelementLen,mskelementLen,reselementLen,
                      status,srcmajor,OC_ALL);
  
  }
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AnyDim                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AnyDim : public ReduceDim_<int, int> {
public :
  AnyDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<int, int>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    if (resBlock[MBLK_STA] == IN_CORE)
       blockScatterOr(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
    else {
         PFILE* Pfile = (PFILE*) resBase;
         oc_blockScatterlogical(Pfile,resBlock,resExts,rank,
                       (int*)buf,reselementLen,srcelementLen,OC_ANY);
        }

  }

  void opZero(const int resOffset) {
   if (status == IN_CORE)
    blockZero(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    if (status == IN_CORE)
     blockVecOr(resBase + resOffset, resSteps,
		srcBase + srcOffset, srcSteps, srcexts, resRank) ;
    else  oc_blockVeclogical(srcBase,resBase,resOffset,srcOffset,
                     srcexts,resexts,srcSteps,resSteps,
                     srcRank,resRank,srcvol,resvol,srcelementLen,reselementLen,
                     status,srcmajor,OC_ANY);  
  }
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AnyDimMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AnyDimMsk : public ReduceDim2_<int, int, int> {
public :
  AnyDimMsk(const DAD* _res,
            const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<int, int, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
//xh
    if (resBlock[MBLK_STA] == IN_CORE) 
      blockScatterOr(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
    else {
      PFILE* Pfile = (PFILE*)resBase;
      oc_blockScatterlogical(Pfile,resBlock,resExts,rank,(int*)buf,reselementLen,srcelementLen,OC_ANY);
    }
  }

  void opZero(const int resOffset) {
//xh
   if (status == IN_CORE)
      blockZero(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset,
                const int mskOffset) {
//xh
   if (status == IN_CORE)
       blockVecOr(resBase + resOffset, resSteps,
		src1Base + srcOffset, src1Steps,
		src2Base + mskOffset, src2Steps, srcexts, resRank) ;
   else
       oc_blockVeclogicalMsk(src1Base,src2Base,resBase,
                      srcOffset,mskOffset,resOffset,
                      srcexts,resexts, 
                      src1Steps,src2Steps,resSteps,
                      srcRank,resRank,
                      srcvol,resvol,
                      srcelementLen,mskelementLen,reselementLen,
                      status,srcmajor,OC_ANY);
  
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class CountDim                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CountDim : public ReduceDim_<int, int> {
public :
  CountDim(const DAD* _res, const DAD* _src, const int _dim) :
           ReduceDim_<int, int>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    if (resBlock[MBLK_STA] == IN_CORE)
      blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
    else {
          PFILE* Pfile = (PFILE*) resBase;
          oc_blockScatterlogical(Pfile,resBlock,resExts,rank,(int*)buf,reselementLen,srcelementLen,OC_COUNT);
        }

  }

  void opZero(const int resOffset) {
    if (status == IN_CORE) 
      blockZero(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    if (status == IN_CORE)
      blockVecCount(resBase + resOffset, resSteps,
		  srcBase + srcOffset, srcSteps, srcexts, resRank) ;
    else  oc_blockVeclogical(srcBase,resBase,resOffset,srcOffset,
                     srcexts,resexts,srcSteps,resSteps,
                     srcRank,resRank,srcvol,resvol,srcelementLen,reselementLen,
                     status,srcmajor,OC_COUNT); 

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class CountDimMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CountDimMsk : public ReduceDim2_<int, int, int> {
public :
  CountDimMsk(const DAD* _res,
            const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<int, int, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
//xh
    if (resBlock[MBLK_STA] == IN_CORE) 
      blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
    else {
      PFILE* Pfile = (PFILE*)resBase;
      oc_blockScatterlogical(Pfile,resBlock,resExts,rank,(int*)buf,reselementLen,srcelementLen,OC_COUNT);
    }
  }

  void opZero(const int resOffset) {
//xh
   if (status == IN_CORE)
      blockZero(resBase + resOffset, resSteps, resexts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset,
                const int mskOffset) {
//xh
   if (status == IN_CORE)
       blockVecCount(resBase + resOffset, resSteps,
		src1Base + srcOffset, src1Steps,
		src2Base + mskOffset, src2Steps, srcexts, resRank) ;
   else
       oc_blockVeclogicalMsk(src1Base,src2Base,resBase,
                      srcOffset,mskOffset,resOffset,
                      srcexts,resexts, 
                      src1Steps,src2Steps,resSteps,
                      srcRank,resRank,
                      srcvol,resvol,
                      srcelementLen,mskelementLen,reselementLen,
                      status,srcmajor,OC_COUNT);
  
  }
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxlocDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxlocDim : public LocateDim_<T, T> {
public :
  MaxlocDim(const DAD* _res, const DAD* _loc,
            const DAD* _src, const int _dim) :
      LocateDim_<T, T>(_res, _loc, _src, _dim) {}

private :
  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valSteps       = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagSteps       = tagBlock + MBLK_STR ;
    int* exts           = valSteps + rank     ;
    if (valBlock[MBLK_STA] == IN_CORE )
      blockScatterMaxtag(valBase + valOffset, valSteps,
                       tagBase + tagOffset, tagSteps, exts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
    else { 
           PFILE* resPfile = (PFILE*)valBase;
           PFILE* locPfile = (PFILE*)tagBase;
           oc_blockScatterLoc(resPfile,locPfile,valBlock,tagBlock,exts,rank,
                              *((T**) &bufRes),*((int**) &bufTag),
                              srcelementLen,locelementLen,OC_MAXLOC);
         }

  }

  void opZero(const int valOffset, const int tagOffset) {
   if (status == IN_CORE) {
    blockMostneg(valBase + valOffset, valSteps, resexts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, resexts, resRank) ;
   } 
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset,
                const int blockBase, const int blockStep) {
    if (status == IN_CORE)
      blockVecMaxloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   srcBase + srcOffset, srcSteps, blockBase, blockStep,
                   srcexts, resRank) ;
    else { 
        oc_blockVecLoc(srcBase,valBase,tagBase,
                       srcOffset,valOffset,tagOffset,
                       srcexts,resexts,
                       srcSteps,valSteps,tagSteps,
                       srcRank,resRank,
                       srcvol,resvol,
                       srcelementLen,locelementLen,
                       blockBase,blockStep,
                       status,srcmajor,OC_MAXLOC);
       
        }                    
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxlocDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxlocDimMsk : public LocateDim2_<T, T, int> {
public :
  MaxlocDimMsk(const DAD* _res, const DAD* _loc,
               const DAD* _src, const int _dim, const DAD* _msk) :
      LocateDim2_<T, T, int>(_res, _loc, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valSteps       = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagSteps       = tagBlock + MBLK_STR ;
    int* exts           = valSteps + rank     ;
    if (valBlock[MBLK_STA] == IN_CORE)
      blockScatterMaxtag(valBase + valOffset, valSteps,
                       tagBase + tagOffset, tagSteps, exts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
    else {
           PFILE* resPfile = (PFILE*)valBase;
           PFILE* locPfile = (PFILE*)tagBase;
           oc_blockScatterLoc(resPfile,locPfile,valBlock,tagBlock,exts,rank,
                              *((T**) &bufRes),*((int**) &bufTag),
                              srcelementLen,locelementLen,OC_MAXLOC);
         }

  }

  void opZero(const int valOffset, const int tagOffset) {
  if(status == IN_CORE) {
    blockMostneg(valBase + valOffset, valSteps, resexts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, resexts, resRank) ;
   }
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset, const int mskOffset,
                const int blockBase, const int blockStep) {
  if(status == IN_CORE) 
    blockVecMaxloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   src1Base + srcOffset, src1Steps,
                   src2Base + mskOffset, src2Steps,
                   blockBase, blockStep,
                   srcexts, resRank) ;
  else  oc_blockVecLocMsk(src1Base, src2Base, valBase, tagBase,
                          srcOffset, mskOffset, valOffset, tagOffset,
                          srcexts, resexts,
                          src1Steps, src2Steps, valSteps, tagSteps,
                          srcRank, resRank,
                          srcvol, resvol,
                          srcelementLen, mskelementLen,locelementLen,
                          blockBase,blockStep,
                          status,srcmajor,OC_MAXLOC);

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinlocDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinlocDim : public LocateDim_<T, T> {
public :
  MinlocDim(const DAD* _res, const DAD* _loc,
            const DAD* _src, const int _dim) :
      LocateDim_<T, T>(_res, _loc, _src, _dim) {}

private :
  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valSteps       = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagSteps       = tagBlock + MBLK_STR ;
    int* exts           = valSteps + rank     ;
   if (valBlock[MBLK_STA] == IN_CORE)
    blockScatterMintag(valBase + valOffset, valSteps,
                       tagBase + tagOffset, tagSteps, exts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
   else { 
           PFILE*  resPfile = (PFILE*)valBase;
           PFILE*  locPfile = (PFILE*)tagBase;
           oc_blockScatterLoc(resPfile,locPfile,valBlock,tagBlock,exts,rank,
                              *((T**) &bufRes),*((int**) &bufTag),
                              srcelementLen,locelementLen,OC_MINLOC);
          }


  }

  void opZero(const int valOffset, const int tagOffset) {
  if (status == IN_CORE){
    blockMostpos(valBase + valOffset, valSteps, resexts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, resexts, resRank) ;
   }
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset,
                const int blockBase, const int blockStep) {
  if (status == IN_CORE)
    blockVecMinloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   srcBase + srcOffset, srcSteps, blockBase, blockStep,
                   srcexts, resRank) ;
   else {
          oc_blockVecLoc(srcBase,valBase,tagBase,
                       srcOffset,valOffset,tagOffset,
                       srcexts,resexts,
                       srcSteps,valSteps,tagSteps,
                       srcRank,resRank,
                       srcvol,resvol,
                       srcelementLen,locelementLen,
                       blockBase,blockStep,
                       status,srcmajor,OC_MINLOC);
      }

  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinlocDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinlocDimMsk : public LocateDim2_<T, T, int> {
public :
  MinlocDimMsk(const DAD* _res, const DAD* _loc,
               const DAD* _src, const int _dim, const DAD* _msk) :
      LocateDim2_<T, T, int>(_res, _loc, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valSteps       = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagSteps       = tagBlock + MBLK_STR ;
    int* exts           = valSteps + rank     ;
    if (valBlock[MBLK_STA] == IN_CORE)
     blockScatterMintag(valBase + valOffset, valSteps,
                       tagBase + tagOffset, tagSteps, exts, rank,
                     *((T**) &bufRes), *((int**) &bufTag)) ;
    else {
          PFILE* resPfile = (PFILE*)valBase;
          PFILE* locPfile = (PFILE*)tagBase;
          oc_blockScatterLoc(resPfile,locPfile,valBlock,tagBlock,exts,rank,
                              *((T**) &bufRes),*((int**) &bufTag),
                              srcelementLen,locelementLen,OC_MINLOC);
         } 


  }

  void opZero(const int valOffset, const int tagOffset) {
  if (status == IN_CORE){
    blockMostpos(valBase + valOffset, valSteps, resexts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, resexts, resRank) ;
  }
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset, const int mskOffset,
                const int blockBase, const int blockStep) {
  if (status == IN_CORE)
   blockVecMinloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   src1Base + srcOffset, src1Steps,
                   src2Base + mskOffset, src2Steps,
                   blockBase, blockStep,
                   srcexts, resRank) ;
   else   oc_blockVecLocMsk(src1Base, src2Base, valBase, tagBase,
                          srcOffset, mskOffset, valOffset, tagOffset,
                          srcexts, resexts,
                          src1Steps, src2Steps, valSteps, tagSteps,
                          srcRank, resRank,
                          srcvol, resvol,
                          srcelementLen, mskelementLen,locelementLen,
                          blockBase,blockStep,
                          status,srcmajor,OC_MINLOC);


  }
} ;


// ----------------------------------------------------------------------
// Things to do
//
//   `tagBase' argument of LocateDim :; execute' should be `int*'?
//
//   Typed arithmetic reductions should support result of different
//   type to source?

